from collections.abc import Hashable, Iterable, Sequence
from functools import lru_cache
from typing import Any, cast, final

from cognite.client.data_classes import (
    DatapointsList,
    DatapointSubscription,
    DatapointSubscriptionList,
    DataPointSubscriptionWrite,
    DatapointSubscriptionWriteList,
    TimeSeries,
    TimeSeriesList,
    TimeSeriesWrite,
    TimeSeriesWriteList,
)
from cognite.client.data_classes.capabilities import (
    Capability,
    TimeSeriesAcl,
    TimeSeriesSubscriptionsAcl,
)
from cognite.client.exceptions import CogniteAPIError, CogniteNotFoundError
from cognite.client.utils.useful_types import SequenceNotStr

from cognite_toolkit._cdf_tk._parameters import ANY_STR, ANYTHING, ParameterSpec, ParameterSpecSet
from cognite_toolkit._cdf_tk.constants import MAX_TIMESTAMP_MS, MIN_TIMESTAMP_MS
from cognite_toolkit._cdf_tk.exceptions import (
    ToolkitRequiredValueError,
)
from cognite_toolkit._cdf_tk.loaders._base_loaders import ResourceContainerLoader, ResourceLoader
from cognite_toolkit._cdf_tk.resource_classes import TimeSeriesYAML
from cognite_toolkit._cdf_tk.utils.diff_list import diff_list_hashable, diff_list_identifiable, dm_identifier

from .auth_loaders import GroupAllScopedLoader, SecurityCategoryLoader
from .classic_loaders import AssetLoader
from .data_organization_loaders import DataSetsLoader


@final
class TimeSeriesLoader(ResourceContainerLoader[str, TimeSeriesWrite, TimeSeries, TimeSeriesWriteList, TimeSeriesList]):
    item_name = "datapoints"
    folder_name = "timeseries"
    filename_pattern = r"^(?!.*DatapointSubscription$).*"
    resource_cls = TimeSeries
    resource_write_cls = TimeSeriesWrite
    list_cls = TimeSeriesList
    list_write_cls = TimeSeriesWriteList
    yaml_cls = TimeSeriesYAML
    kind = "TimeSeries"
    dependencies = frozenset({DataSetsLoader, GroupAllScopedLoader, AssetLoader})
    _doc_url = "Time-series/operation/postTimeSeries"

    @property
    def display_name(self) -> str:
        return "time series"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[TimeSeriesWrite] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = [TimeSeriesAcl.Action.Read] if read_only else [TimeSeriesAcl.Action.Read, TimeSeriesAcl.Action.Write]

        scope: TimeSeriesAcl.Scope.All | TimeSeriesAcl.Scope.DataSet = TimeSeriesAcl.Scope.All()  # type: ignore[valid-type]
        if items:
            if dataset_ids := {item.data_set_id for item in items if item.data_set_id}:
                scope = TimeSeriesAcl.Scope.DataSet(list(dataset_ids))

        return TimeSeriesAcl(
            actions,
            scope,  # type: ignore[arg-type]
        )

    @classmethod
    def get_id(cls, item: TimeSeries | TimeSeriesWrite | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        if item.external_id is None:
            raise ToolkitRequiredValueError("TimeSeries must have external_id set.")
        return item.external_id

    @classmethod
    def get_internal_id(cls, item: TimeSeries | dict) -> int:
        if isinstance(item, dict):
            return item["id"]
        return item.id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceLoader], Hashable]]:
        if "dataSetExternalId" in item:
            yield DataSetsLoader, item["dataSetExternalId"]
        if "securityCategoryNames" in item:
            for security_category in item["securityCategoryNames"]:
                yield SecurityCategoryLoader, security_category
        if "assetExternalId" in item:
            yield AssetLoader, item["assetExternalId"]

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> TimeSeriesWrite:
        if ds_external_id := resource.pop("dataSetExternalId", None):
            resource["dataSetId"] = self.client.lookup.data_sets.id(ds_external_id, is_dry_run)
        if security_categories_names := resource.pop("securityCategoryNames", []):
            resource["securityCategories"] = self.client.lookup.security_categories.id(
                security_categories_names, is_dry_run
            )
        if asset_external_id := resource.pop("assetExternalId", None):
            resource["assetId"] = self.client.lookup.assets.id(asset_external_id, is_dry_run)
        return TimeSeriesWrite._load(resource)

    def dump_resource(self, resource: TimeSeries, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump()
        if data_set_id := dumped.pop("dataSetId", None):
            dumped["dataSetExternalId"] = self.client.lookup.data_sets.external_id(data_set_id)
        if security_categories := dumped.pop("securityCategories", []):
            dumped["securityCategoryNames"] = self.client.lookup.security_categories.external_id(security_categories)
        if asset_id := dumped.pop("assetId", None):
            dumped["assetExternalId"] = self.client.lookup.assets.external_id(asset_id)
        return dumped

    def create(self, items: TimeSeriesWriteList) -> TimeSeriesList:
        return self.client.time_series.create(items)

    def retrieve(self, ids: SequenceNotStr[str | int]) -> TimeSeriesList:
        internal_ids, external_ids = self._split_ids(ids)
        return self.client.time_series.retrieve_multiple(
            ids=internal_ids, external_ids=external_ids, ignore_unknown_ids=True
        )

    def update(self, items: TimeSeriesWriteList) -> TimeSeriesList:
        return self.client.time_series.update(items, mode="replace")

    def delete(self, ids: SequenceNotStr[str | int]) -> int:
        existing = self.retrieve(ids)
        if existing:
            self.client.time_series.delete(id=existing.as_ids(), ignore_unknown_ids=True)
        return len(existing)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[TimeSeries]:
        return iter(
            self.client.time_series(data_set_external_ids=[data_set_external_id] if data_set_external_id else None)
        )

    def count(self, ids: str | dict[str, Any] | SequenceNotStr[str | dict[str, Any]] | None) -> int:
        datapoints = cast(
            DatapointsList,
            self.client.time_series.data.retrieve(
                external_id=cast(SequenceNotStr[str], ids),
                start=MIN_TIMESTAMP_MS,
                end=MAX_TIMESTAMP_MS + 1,
                aggregates="count",
                granularity="1000d",
                ignore_unknown_ids=True,
            ),
        )
        return sum(sum(data.count or []) for data in datapoints)

    def drop_data(self, ids: SequenceNotStr[str] | None) -> int:
        count = self.count(ids)
        existing = self.client.time_series.retrieve_multiple(
            external_ids=cast(SequenceNotStr[str], ids), ignore_unknown_ids=True
        ).as_external_ids()
        for external_id in existing:
            self.client.time_series.data.delete_range(
                external_id=external_id, start=MIN_TIMESTAMP_MS, end=MAX_TIMESTAMP_MS + 1
            )
        return count

    @classmethod
    @lru_cache(maxsize=1)
    def get_write_cls_parameter_spec(cls) -> ParameterSpecSet:
        spec = super().get_write_cls_parameter_spec()
        # Added by toolkit
        spec.add(ParameterSpec(("dataSetExternalId",), frozenset({"str"}), is_required=False, _is_nullable=False))

        spec.add(ParameterSpec(("securityCategoryNames",), frozenset({"list"}), is_required=False, _is_nullable=False))

        spec.add(
            ParameterSpec(("securityCategoryNames", ANY_STR), frozenset({"str"}), is_required=False, _is_nullable=False)
        )
        spec.add(ParameterSpec(("assetExternalId",), frozenset({"str"}), is_required=False, _is_nullable=False))
        spec.discard(ParameterSpec(("assetId",), frozenset({"int"}), is_required=False, _is_nullable=False))
        return spec


@final
class DatapointSubscriptionLoader(
    ResourceLoader[
        str,
        DataPointSubscriptionWrite,
        DatapointSubscription,
        DatapointSubscriptionWriteList,
        DatapointSubscriptionList,
    ]
):
    folder_name = "timeseries"
    filename_pattern = r"^.*DatapointSubscription$"  # Matches all yaml files who end with *DatapointSubscription.
    resource_cls = DatapointSubscription
    resource_write_cls = DataPointSubscriptionWrite
    list_cls = DatapointSubscriptionList
    list_write_cls = DatapointSubscriptionWriteList
    kind = "DatapointSubscription"
    _doc_url = "Data-point-subscriptions/operation/postSubscriptions"
    dependencies = frozenset(
        {
            TimeSeriesLoader,
            GroupAllScopedLoader,
        }
    )

    @property
    def display_name(self) -> str:
        return "timeseries subscriptions"

    @classmethod
    def get_id(cls, item: DataPointSubscriptionWrite | DatapointSubscription | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        return item.external_id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    @classmethod
    @lru_cache(maxsize=1)
    def get_write_cls_parameter_spec(cls) -> ParameterSpecSet:
        spec = super().get_write_cls_parameter_spec()
        # Added by toolkit
        spec.add(ParameterSpec(("dataSetExternalId",), frozenset({"str"}), is_required=False, _is_nullable=False))
        # The Filter class in the SDK class View implementation is deviating from the API.
        # So we need to modify the spec to match the API.
        parameter_path = ("filter",)
        length = len(parameter_path)
        for item in spec:
            if len(item.path) >= length + 1 and item.path[:length] == parameter_path[:length]:
                # Add extra ANY_STR layer
                # The spec class is immutable, so we use this trick to modify it.
                object.__setattr__(item, "path", item.path[:length] + (ANY_STR,) + item.path[length:])
        spec.add(ParameterSpec(("filter", ANYTHING), frozenset({"dict"}), is_required=False, _is_nullable=False))
        return spec

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceLoader], Hashable]]:
        if "dataSetExternalId" in item:
            yield DataSetsLoader, item["dataSetExternalId"]
        for timeseries_id in item.get("timeSeriesIds", []):
            yield TimeSeriesLoader, timeseries_id

    @classmethod
    def get_required_capability(
        cls, items: Sequence[DataPointSubscriptionWrite] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [TimeSeriesSubscriptionsAcl.Action.Read]
            if read_only
            else [TimeSeriesSubscriptionsAcl.Action.Read, TimeSeriesSubscriptionsAcl.Action.Write]
        )

        scope: TimeSeriesSubscriptionsAcl.Scope.All | TimeSeriesSubscriptionsAcl.Scope.DataSet = (  # type: ignore[valid-type]
            TimeSeriesSubscriptionsAcl.Scope.All()
        )
        if items:
            if data_set_ids := {item.data_set_id for item in items if item.data_set_id}:
                scope = TimeSeriesSubscriptionsAcl.Scope.DataSet(list(data_set_ids))

        return TimeSeriesSubscriptionsAcl(
            actions,
            scope,  # type: ignore[arg-type]
        )

    def create(self, items: DatapointSubscriptionWriteList) -> DatapointSubscriptionList:
        created = DatapointSubscriptionList([])
        for item in items:
            created.append(self.client.time_series.subscriptions.create(item))
        return created

    def retrieve(self, ids: SequenceNotStr[str]) -> DatapointSubscriptionList:
        items = DatapointSubscriptionList([])
        for id_ in ids:
            retrieved = self.client.time_series.subscriptions.retrieve(id_)
            if retrieved:
                items.append(retrieved)
        return items

    def update(self, items: DatapointSubscriptionWriteList) -> DatapointSubscriptionList:
        updated = DatapointSubscriptionList([])
        for item in items:
            # There are two versions of a TimeSeries Subscription, one selects timeseries based filter
            # and the other selects timeseries based on timeSeriesIds. If we use mode='replace', we try
            # to set timeSeriesIds to an empty list, while the filter is set. This will result in an error.
            update = self.client.time_series.subscriptions.update(item, mode="replace_ignore_null")
            updated.append(update)

        return updated

    def delete(self, ids: SequenceNotStr[str]) -> int:
        try:
            self.client.time_series.subscriptions.delete(ids)
        except (CogniteAPIError, CogniteNotFoundError) as e:
            non_existing = set(e.failed or [])
            if existing := [id_ for id_ in ids if id_ not in non_existing]:
                self.client.time_series.subscriptions.delete(existing)
            return len(existing)
        else:
            # All deleted successfully
            return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[DatapointSubscription]:
        return iter(self.client.time_series.subscriptions)

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> DataPointSubscriptionWrite:
        if ds_external_id := resource.pop("dataSetExternalId", None):
            resource["dataSetId"] = self.client.lookup.data_sets.id(ds_external_id, is_dry_run)
        return DataPointSubscriptionWrite._load(resource)

    def dump_resource(self, resource: DatapointSubscription, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump()
        local = local or {}
        if data_set_id := dumped.pop("dataSetId", None):
            dumped["dataSetExternalId"] = self.client.lookup.data_sets.external_id(data_set_id)
        if "timeSeriesIds" not in dumped:
            return dumped
        # Sorting the timeSeriesIds in the local order
        # Sorting in the same order as the local file.
        ts_order_by_id = {ts_id: no for no, ts_id in enumerate(local.get("timeSeriesIds", []))}
        end_of_list = len(ts_order_by_id)
        dumped["timeSeriesIds"] = sorted(
            dumped["timeSeriesIds"], key=lambda ts_id: ts_order_by_id.get(ts_id, end_of_list)
        )
        return dumped

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path[0] == "filter" or json_path == ("timeSeriesIds",):
            return diff_list_hashable(local, cdf)
        elif json_path == ("instanceIds",):
            return diff_list_identifiable(local, cdf, get_identifier=dm_identifier)
        return super().diff_list(local, cdf, json_path)



from __future__ import annotations

from collections.abc import Iterator, Sequence
from typing import overload

from cognite.client._api_client import APIClient

from cognite_toolkit._cdf_tk.client.api.robotics.utlis import tmp_disable_gzip
from cognite_toolkit._cdf_tk.client.data_classes.robotics import Robot, RobotList, RobotWrite, _RobotUpdate


class RobotsAPI(APIClient):
    _RESOURCE_PATH = "/robotics/robots"

    @overload
    def __call__(self) -> Iterator[Robot]: ...

    @overload
    def __call__(self, chunk_size: int) -> Iterator[RobotList]: ...

    def __call__(self, chunk_size: int | None = None) -> Iterator[Robot] | Iterator[RobotList]:
        """Iterate over robots.

        Args:
            chunk_size: The number of robots to return in each chunk. None will return all robots.

        Yields:
            Robot or RobotList

        """
        return self._list_generator(method="GET", resource_cls=Robot, list_cls=RobotList, chunk_size=chunk_size)

    def __iter__(self) -> Iterator[Robot]:
        return self.__call__()

    @overload
    def create(self, robot: RobotWrite) -> Robot: ...

    @overload
    def create(self, robot: Sequence[RobotWrite]) -> RobotList: ...

    def create(self, robot: RobotWrite | Sequence[RobotWrite]) -> Robot | RobotList:
        """Create a new robot.

        Args:
            robot: RobotWrite or list of RobotWrite.

        Returns:
            Robot object.

        """
        with tmp_disable_gzip():
            return self._create_multiple(
                list_cls=RobotList,
                resource_cls=Robot,
                items=robot,
                input_resource_cls=RobotWrite,
            )

    def retrieve(self, data_set_id: int | Sequence[int]) -> RobotList:
        """Retrieve a robot.

        Args:
            data_set_id: Data set id of the robot.

        Returns:
            Robot object.

        """
        body = self._create_body(data_set_id)
        res = self._post(url_path=self._RESOURCE_PATH + "/byids", json=body)
        return RobotList._load(res.json()["items"], cognite_client=self._cognite_client)

    @staticmethod
    def _create_body(data_set_id: int | Sequence[int]) -> dict:
        ids = [data_set_id] if isinstance(data_set_id, int) else data_set_id
        body = {"items": [{"dataSetId": external_id} for external_id in ids]}
        return body

    @overload
    def update(self, robot: RobotWrite) -> Robot: ...

    @overload
    def update(self, robot: Sequence[RobotWrite]) -> RobotList: ...

    def update(self, robot: RobotWrite | Sequence[RobotWrite]) -> Robot | RobotList:
        """Update a robot.

        Args:
            robot: RobotWrite or list of RobotWrite.

        Returns:
            Robot object.

        """
        is_single = False
        if isinstance(robot, RobotWrite):
            robots = [robot]
            is_single = True
        elif isinstance(robot, Sequence):
            robots = list(robot)
        else:
            raise ValueError("robot must be a RobotWrite or a list of RobotWrite")

        property_spec = _RobotUpdate._get_update_properties()
        update = [
            {"dataSetId": r.data_set_id, **self._convert_resource_to_patch_object(r, property_spec)} for r in robots
        ]
        res = self._post(url_path=self._RESOURCE_PATH + "/update", json={"items": update})
        loaded = RobotList._load(res.json()["items"], cognite_client=self._cognite_client)
        return loaded[0] if is_single else loaded

    def delete(self, data_set_id: int | Sequence[int]) -> None:
        """Delete a robot.

        Args:
            data_set_id: Data set id of the robot.

        Returns:
            None

        """
        body = self._create_body(data_set_id)
        self._post(url_path=self._RESOURCE_PATH + "/delete", json=body)

    def list(self) -> RobotList:
        """List robots.

        Returns:
            RobotList

        """
        with tmp_disable_gzip():
            return self._list(method="GET", resource_cls=Robot, list_cls=RobotList)



from pathlib import Path

DATA_FOLDER = Path(__file__).resolve().parent

DESCRIPTIONS_FOLDER = DATA_FOLDER / "describe_data"
AUTH_DATA = DATA_FOLDER / "auth_data"
PROJECT_NO_COGNITE_MODULES = DATA_FOLDER / "project_no_cognite_modules"
RESOURCES_WITH_ENVIRONMENT_VARIABLES = DATA_FOLDER / "resources_with_environment_variables"
PROJECT_WITH_DUPLICATES = DATA_FOLDER / "project_with_duplicates"
PROJECT_FOR_TEST = DATA_FOLDER / "project_for_test"
LOAD_DATA = DATA_FOLDER / "load_data"
RUN_DATA = DATA_FOLDER / "run_data"
TRANSFORMATION_CLI = DATA_FOLDER / "transformation_cli"
PROJECT_WITH_BAD_MODULES = DATA_FOLDER / "project_with_bad_modules"
NAUGHTY_PROJECT = DATA_FOLDER / "naughty_project"

BUILD_GROUP_WITH_UNKNOWN_ACL = DATA_FOLDER / "build_group_with_unknown_acl"
COMPLETE_ORG = DATA_FOLDER / "complete_org"
COMPLETE_ORG_ALPHA_FLAGS = DATA_FOLDER / "complete_org_alpha_flags"
COMPLETE_ORG_ONLY_IDENTIFIER = DATA_FOLDER / "complete_org_only_identifier"
CDF_TOML_DATA = DATA_FOLDER / "cdf_toml_data"
STRONGLY_COUPLED_MODEL = DATA_FOLDER / "strongly_coupled_model"

__all__ = [
    "AUTH_DATA",
    "BUILD_GROUP_WITH_UNKNOWN_ACL",
    "CDF_TOML_DATA",
    "COMPLETE_ORG",
    "DATA_FOLDER",
    "DESCRIPTIONS_FOLDER",
    "LOAD_DATA",
    "PROJECT_FOR_TEST",
    "PROJECT_NO_COGNITE_MODULES",
    "PROJECT_WITH_BAD_MODULES",
    "PROJECT_WITH_DUPLICATES",
    "RESOURCES_WITH_ENVIRONMENT_VARIABLES",
    "RUN_DATA",
    "STRONGLY_COUPLED_MODEL",
    "TRANSFORMATION_CLI",
]



from abc import abstractmethod
from functools import lru_cache

import questionary
from cognite.client.data_classes import (
    Asset,
    AssetFilter,
    AssetList,
    DataSet,
    DataSetList,
    EventFilter,
    FileMetadataFilter,
    TimeSeriesFilter,
)

from cognite_toolkit._cdf_tk.client import ToolkitClient


class AssetCentricInteractiveSelect:
    def __init__(self, client: ToolkitClient) -> None:
        self.client = client

    @lru_cache
    def aggregate_count(self, hierarchies: tuple[str, ...], data_sets: tuple[str, ...]) -> int:
        return self._aggregate_count(list(hierarchies), list(data_sets))

    @abstractmethod
    def _aggregate_count(self, hierarchies: list[str], data_sets: list[str]) -> int:
        raise NotImplementedError()

    @lru_cache(maxsize=1)
    def _get_available_data_sets(self) -> DataSetList:
        return self.client.data_sets.list(limit=-1)

    @lru_cache(maxsize=1)
    def _get_available_hierarchies(self) -> AssetList:
        return self.client.assets.list(root=True, limit=-1)

    def _create_choice(self, item: Asset | DataSet) -> tuple[questionary.Choice, int]:
        """Create a questionary choice for the given item."""

        if isinstance(item, DataSet):
            if item.external_id is None:
                raise ValueError(f"Missing external ID for DataSet {item.id}")
            item_count = self.aggregate_count(tuple(), (item.external_id,))
        elif isinstance(item, Asset):
            if item.external_id is None:
                raise ValueError(f"Missing external ID for Asset {item.id}")
            item_count = self.aggregate_count((item.external_id,), tuple())
        else:
            raise TypeError(f"Unsupported item type: {type(item)}")

        return questionary.Choice(
            title=f"{item.name} ({item.external_id}) [{item_count:,}]"
            if item.name != item.external_id
            else f"({item.external_id}) [{item_count:,}]",
            value=item.external_id,
        ), item_count

    def interactive_select_hierarchy_datasets(self) -> tuple[list[str], list[str]]:
        """Interactively select hierarchies and data sets to dump."""
        hierarchies: set[str] = set()
        data_sets: set[str] = set()
        while True:
            selected: list[str] = []
            if hierarchies:
                selected.append(f"Selected hierarchies: {sorted(hierarchies)}")
            else:
                selected.append("No hierarchy selected.")
            if data_sets:
                selected.append(f"Selected data sets: {sorted(data_sets)}")
            else:
                selected.append("No data set selected.")
            selected_str = "\n".join(selected)
            what = questionary.select(
                f"\n{selected_str}\nSelect a hierarchy or data set to dump",
                choices=["Hierarchy", "Data Set", "Done", "Abort"],
            ).ask()

            if what == "Done":
                break
            elif what == "Abort":
                return [], []
            elif what == "Hierarchy":
                options = [asset for asset in self._get_available_hierarchies() if asset.external_id not in hierarchies]
                selected_hierarchy = self._select(what, options)
                if selected_hierarchy:
                    hierarchies.update(selected_hierarchy)
                else:
                    print("No hierarchy selected.")
            elif what == "Data Set":
                available_data_sets = [
                    data_set for data_set in self._get_available_data_sets() if data_set.external_id not in data_sets
                ]
                selected_data_set = self._select(what, available_data_sets)
                if selected_data_set:
                    data_sets.update(selected_data_set)
                else:
                    print("No data set selected.")
        return list(hierarchies), list(data_sets)

    def _select(self, what: str, options: list[Asset] | list[DataSet]) -> str | None:
        return questionary.checkbox(
            f"Select a {what} listed as 'name (external_id) [count]'",
            choices=[
                choice
                for choice, count in (
                    # MyPy does not seem to understand that item is Asset | DataSet
                    self._create_choice(item)  # type: ignore[arg-type]
                    for item in sorted(options, key=lambda x: x.name or x.external_id)
                )
                if count > 0
            ],
        ).ask()


class AssetInteractiveSelect(AssetCentricInteractiveSelect):
    def _aggregate_count(self, hierarchies: list[str], data_sets: list[str]) -> int:
        return self.client.assets.aggregate_count(
            filter=AssetFilter(
                data_set_ids=[{"externalId": item} for item in data_sets] or None,
                asset_subtree_ids=[{"externalId": item} for item in hierarchies] or None,
            )
        )


class FileMetadataInteractiveSelect(AssetCentricInteractiveSelect):
    def _aggregate_count(self, hierarchies: list[str], data_sets: list[str]) -> int:
        result = self.client.files.aggregate(
            filter=FileMetadataFilter(
                data_set_ids=[{"externalId": item} for item in data_sets] or None,
                asset_subtree_ids=[{"externalId": item} for item in hierarchies] or None,
            )
        )
        return result[0].count if result else 0


class TimeSeriesInteractiveSelect(AssetCentricInteractiveSelect):
    def _aggregate_count(self, hierarchies: list[str], data_sets: list[str]) -> int:
        return self.client.time_series.aggregate_count(
            filter=TimeSeriesFilter(
                data_set_ids=[{"externalId": item} for item in data_sets] or None,
                asset_subtree_ids=[{"externalId": item} for item in hierarchies] or None,
            )
        )


class EventInteractiveSelect(AssetCentricInteractiveSelect):
    def _aggregate_count(self, hierarchies: list[str], data_sets: list[str]) -> int:
        return self.client.events.aggregate_count(
            filter=EventFilter(
                data_set_ids=[{"externalId": item} for item in data_sets] or None,
                asset_subtree_ids=[{"externalId": item} for item in hierarchies] or None,
            )
        )



import pytest

from cognite_toolkit._cdf_tk.resource_classes import SecurityCategoriesYAML
from tests.test_unit.utils import find_resources


class TestSecuritytCategoriesYAML:
    @pytest.mark.parametrize("data", list(find_resources("SecurityCategory")))
    def test_load_valid_security_categories(self, data: dict[str, object]) -> None:
        loaded = SecurityCategoriesYAML.model_validate(data)

        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data



from abc import ABC, abstractmethod
from collections.abc import Hashable, Iterable, Iterator
from functools import cached_property
from pathlib import Path
from typing import Generic, cast

import questionary
import typer
from cognite.client import data_modeling as dm
from cognite.client.data_classes import (
    Group,
    GroupList,
    TransformationList,
    TransformationNotificationList,
    TransformationScheduleList,
)
from cognite.client.data_classes._base import (
    CogniteResourceList,
)
from cognite.client.data_classes.data_modeling import DataModelId
from cognite.client.data_classes.workflows import (
    Workflow,
    WorkflowList,
    WorkflowTriggerList,
    WorkflowVersion,
    WorkflowVersionId,
    WorkflowVersionList,
)
from cognite.client.exceptions import CogniteAPIError
from questionary import Choice
from rich import print
from rich.panel import Panel

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.location_filters import LocationFilterList
from cognite_toolkit._cdf_tk.exceptions import (
    ResourceRetrievalError,
    ToolkitMissingResourceError,
    ToolkitResourceMissingError,
    ToolkitValueError,
)
from cognite_toolkit._cdf_tk.loaders import (
    ContainerLoader,
    DataModelLoader,
    GroupLoader,
    LocationFilterLoader,
    NodeLoader,
    ResourceLoader,
    SpaceLoader,
    TransformationLoader,
    TransformationNotificationLoader,
    TransformationScheduleLoader,
    ViewLoader,
    WorkflowLoader,
    WorkflowTriggerLoader,
    WorkflowVersionLoader,
)
from cognite_toolkit._cdf_tk.loaders._base_loaders import T_ID
from cognite_toolkit._cdf_tk.tk_warnings import FileExistsWarning, MediumSeverityWarning
from cognite_toolkit._cdf_tk.utils import humanize_collection
from cognite_toolkit._cdf_tk.utils.file import safe_rmtree, safe_write, yaml_safe_dump

from ._base import ToolkitCommand


class ResourceFinder(Iterable, ABC, Generic[T_ID]):
    def __init__(self, client: ToolkitClient, identifier: T_ID | None = None):
        self.client = client
        self.identifier = identifier

    def _selected(self) -> T_ID:
        return self.identifier or self._interactive_select()

    @abstractmethod
    def __iter__(self) -> Iterator[tuple[list[Hashable], CogniteResourceList | None, ResourceLoader, None | str]]:
        raise NotImplementedError

    @abstractmethod
    def _interactive_select(self) -> T_ID:
        raise NotImplementedError

    # Can be implemented in subclasses
    def update(self, resources: CogniteResourceList) -> None: ...


class DataModelFinder(ResourceFinder[DataModelId]):
    def __init__(self, client: ToolkitClient, identifier: DataModelId | None = None, include_global: bool = False):
        super().__init__(client, identifier)
        self._include_global = include_global
        self.data_model: dm.DataModel[dm.ViewId] | None = None
        self.view_ids: set[dm.ViewId] = set()
        self.container_ids: set[dm.ContainerId] = set()
        self.space_ids: set[str] = set()

    def _interactive_select(self) -> DataModelId:
        data_model_ids = self.client.data_modeling.data_models.list(
            all_versions=False, limit=-1, include_global=False
        ).as_ids()
        available_spaces = sorted({model.space for model in data_model_ids})
        if not available_spaces:
            raise ToolkitMissingResourceError("No data models found")
        if len(available_spaces) == 1:
            selected_space = available_spaces[0]
        else:
            selected_space = questionary.select("In which space is your data model located?", available_spaces).ask()
        data_model_ids = sorted(
            [model for model in data_model_ids if model.space == selected_space], key=lambda model: model.as_tuple()
        )

        selected_data_model: DataModelId = questionary.select(
            "Which data model would you like to dump?",
            [
                Choice(f"{model_id!r}", value=model_id)
                for model_id in sorted(data_model_ids, key=lambda model: model.as_tuple())
            ],
        ).ask()

        retrieved_models = self.client.data_modeling.data_models.retrieve(
            (selected_data_model.space, selected_data_model.external_id), inline_views=False
        )
        if not retrieved_models:
            # This happens if the data model is removed after the list call above.
            raise ToolkitMissingResourceError(f"Data model {selected_data_model} not found")
        if len(retrieved_models) == 1:
            self.data_model = retrieved_models[0]
            return selected_data_model
        models_by_version = {model.version: model for model in retrieved_models if model.version is not None}
        if len(models_by_version) == 1:
            self.data_model = retrieved_models[0]
            return selected_data_model
        if not questionary.confirm(
            f"Would you like to select a different version than {selected_data_model.version} of the data model",
            default=False,
        ).ask():
            self.data_model = models_by_version[cast(str, selected_data_model.version)]
            return selected_data_model

        selected_model = questionary.select(
            "Which version would you like to dump?",
            [Choice(f"{version}", value=model) for version, model in models_by_version.items()],
        ).ask()
        self.data_model = models_by_version[selected_model]
        return self.data_model.as_id()

    def update(self, resources: CogniteResourceList) -> None:
        if isinstance(resources, dm.DataModelList):
            self.view_ids |= {
                view.as_id() if isinstance(view, dm.View) else view for item in resources for view in item.views
            }
        elif isinstance(resources, dm.ViewList):
            self.container_ids |= resources.referenced_containers()
        elif isinstance(resources, dm.SpaceList):
            return
        self.space_ids |= {item.space for item in resources}

    def __iter__(self) -> Iterator[tuple[list[Hashable], CogniteResourceList | None, ResourceLoader, None | str]]:
        self.identifier = self._selected()
        model_loader = DataModelLoader.create_loader(self.client)
        if self.data_model:
            is_global_model = self.data_model.is_global
            yield [], dm.DataModelList([self.data_model]), model_loader, None
        else:
            model_list = model_loader.retrieve([self.identifier])
            if not model_list:
                raise ToolkitResourceMissingError(f"Data model {self.identifier} not found", str(self.identifier))
            is_global_model = model_list[0].is_global
            yield [], model_list, model_loader, None
        if self._include_global or is_global_model:
            yield list(self.view_ids), None, ViewLoader.create_loader(self.client), "views"
            yield list(self.container_ids), None, ContainerLoader.create_loader(self.client), "containers"
            yield list(self.space_ids), None, SpaceLoader.create_loader(self.client), None
        else:
            view_loader = ViewLoader(self.client, None, None, topological_sort_implements=True)
            views = dm.ViewList([view for view in view_loader.retrieve(list(self.view_ids)) if not view.is_global])
            yield [], views, view_loader, "views"
            container_loader = ContainerLoader.create_loader(self.client)
            containers = dm.ContainerList(
                [
                    container
                    for container in container_loader.retrieve(list(self.container_ids))
                    if not container.is_global
                ]
            )
            yield [], containers, container_loader, "containers"

            space_loader = SpaceLoader.create_loader(self.client)
            spaces = dm.SpaceList(
                [space for space in space_loader.retrieve(list(self.space_ids)) if not space.is_global]
            )
            yield [], spaces, space_loader, None


class WorkflowFinder(ResourceFinder[WorkflowVersionId]):
    def __init__(self, client: ToolkitClient, identifier: WorkflowVersionId | None = None):
        super().__init__(client, identifier)
        self._workflow: Workflow | None = None
        self._workflow_version: WorkflowVersion | None = None

    def _interactive_select(self) -> WorkflowVersionId:
        workflows = self.client.workflows.list(limit=-1)
        if not workflows:
            raise ToolkitMissingResourceError("No workflows found")
        selected_workflow_id: str = questionary.select(
            "Which workflow would you like to dump?",
            [Choice(workflow_id, value=workflow_id) for workflow_id in workflows.as_external_ids()],
        ).ask()
        for workflow in workflows:
            if workflow.external_id == selected_workflow_id:
                self._workflow = workflow
                break

        versions = self.client.workflows.versions.list(selected_workflow_id, limit=-1)
        if len(versions) == 0:
            raise ToolkitMissingResourceError(f"No versions found for workflow {selected_workflow_id}")
        if len(versions) == 1:
            self._workflow_version = versions[0]
            return self._workflow_version.as_id()

        selected_version: WorkflowVersionId = questionary.select(
            "Which version would you like to dump?",
            [Choice(f"{version!r}", value=version) for version in versions.as_ids()],
        ).ask()
        for version in versions:
            if version.version == selected_version.version:
                self._workflow_version = version
                break
        return selected_version

    def __iter__(self) -> Iterator[tuple[list[Hashable], CogniteResourceList | None, ResourceLoader, None | str]]:
        self.identifier = self._selected()
        if self._workflow:
            yield [], WorkflowList([self._workflow]), WorkflowLoader.create_loader(self.client), None
        else:
            yield [self.identifier.workflow_external_id], None, WorkflowLoader.create_loader(self.client), None
        if self._workflow_version:
            yield (
                [],
                WorkflowVersionList([self._workflow_version]),
                WorkflowVersionLoader.create_loader(self.client),
                None,
            )
        else:
            yield [self.identifier], None, WorkflowVersionLoader.create_loader(self.client), None
        trigger_loader = WorkflowTriggerLoader.create_loader(self.client)
        trigger_list = WorkflowTriggerList(
            list(trigger_loader.iterate(parent_ids=[self.identifier.workflow_external_id]))
        )
        yield [], trigger_list, trigger_loader, None


class TransformationFinder(ResourceFinder[tuple[str, ...]]):
    def __init__(self, client: ToolkitClient, identifier: tuple[str, ...] | None = None):
        super().__init__(client, identifier)
        self.transformations: TransformationList | None = None

    def _interactive_select(self) -> tuple[str, ...]:
        self.transformations = self.client.transformations.list(limit=-1)
        if self.transformations and not any(transformation.external_id for transformation in self.transformations):
            raise ToolkitValueError(
                "ExternalID is required for dumping transformations. "
                f"Found {len(self.transformations)} transformations with only internal IDs."
            )
        elif not self.transformations:
            raise ToolkitMissingResourceError("No transformations found")

        choices = [
            Choice(f"{transformation.name} ({transformation.external_id})", value=transformation.external_id)
            for transformation in sorted(self.transformations, key=lambda t: t.name or "")
            if transformation.external_id
        ]

        selected_transformation_ids: tuple[str, ...] | None = questionary.checkbox(
            "Which transformation(s) would you like to dump?",
            choices=choices,
        ).ask()
        if selected_transformation_ids is None:
            raise ToolkitValueError("No transformations selected for dumping.")
        return tuple(selected_transformation_ids)

    def __iter__(self) -> Iterator[tuple[list[Hashable], CogniteResourceList | None, ResourceLoader, None | str]]:
        self.identifier = self._selected()
        if self.transformations:
            yield (
                [],
                TransformationList([t for t in self.transformations if t.external_id in self.identifier]),
                TransformationLoader.create_loader(self.client),
                None,
            )
        else:
            yield list(self.identifier), None, TransformationLoader.create_loader(self.client), None

        schedule_loader = TransformationScheduleLoader.create_loader(self.client)
        schedule_list = TransformationScheduleList(schedule_loader.iterate(parent_ids=list(self.identifier)))
        yield [], schedule_list, schedule_loader, None
        notification_loader = TransformationNotificationLoader.create_loader(self.client)
        notification_list = TransformationNotificationList(
            notification_loader.iterate(parent_ids=list(self.identifier))
        )
        yield [], notification_list, notification_loader, None


class GroupFinder(ResourceFinder[str]):
    def __init__(self, client: ToolkitClient, identifier: str | None = None):
        super().__init__(client, identifier)
        self.group: Group | None = None

    def _interactive_select(self) -> str:
        groups = self.client.iam.groups.list(all=True)
        if not groups:
            raise ToolkitMissingResourceError("No groups found")
        group_names = [group.name for group in groups]
        selected_group_name: str = questionary.select(
            "Which group would you like to dump?",
            [Choice(group, value=group) for group in group_names],
        ).ask()
        for group in groups:
            if group.name == selected_group_name:
                self.group = group
                break
        return selected_group_name

    def __iter__(self) -> Iterator[tuple[list[Hashable], CogniteResourceList | None, ResourceLoader, None | str]]:
        self.identifier = self._selected()
        if self.group:
            yield [], GroupList([self.group]), GroupLoader.create_loader(self.client), None
        else:
            yield [self.identifier], None, GroupLoader.create_loader(self.client), None


class NodeFinder(ResourceFinder[dm.ViewId]):
    def __init__(self, client: ToolkitClient, identifier: dm.ViewId | None = None):
        super().__init__(client, identifier)
        self.is_interactive = False

    def _interactive_select(self) -> dm.ViewId:
        self.is_interactive = True
        spaces = self.client.data_modeling.spaces.list(limit=-1)
        if not spaces:
            raise ToolkitMissingResourceError("No spaces found")
        selected_space: str = questionary.select(
            "In which space is your node property view located?", [space.space for space in spaces]
        ).ask()

        views = self.client.data_modeling.views.list(space=selected_space, limit=-1, all_versions=False)
        if not views:
            raise ToolkitMissingResourceError(f"No views found in {selected_space}")
        if len(views) == 1:
            return views[0].as_id()
        selected_view_id: dm.ViewId = questionary.select(
            "Which node property view would you like to dump?",
            [Choice(repr(view), value=view) for view in views.as_ids()],
        ).ask()
        return selected_view_id

    def __iter__(self) -> Iterator[tuple[list[Hashable], CogniteResourceList | None, ResourceLoader, None | str]]:
        self.identifier = self._selected()
        loader = NodeLoader(self.client, None, None, self.identifier)
        if self.is_interactive:
            count = self.client.data_modeling.instances.aggregate(
                self.identifier, dm.aggregations.Count("externalId"), instance_type="node"
            ).value
            if count == 0 or count is None:
                raise ToolkitMissingResourceError(f"No nodes found in {self.identifier}")
            elif count > 50:
                if not questionary.confirm(
                    f"Are you sure you want to dump {count} nodes? This may take a while.",
                    default=False,
                ).ask():
                    typer.Exit(0)
        nodes = dm.NodeList[dm.Node](list(loader.iterate()))
        yield [], nodes, loader, None


class LocationFilterFinder(ResourceFinder[tuple[str, ...]]):
    @cached_property
    def all_filters(self) -> LocationFilterList:
        return self.client.location_filters.list()

    def _interactive_select(self) -> tuple[str, ...]:
        filters = self.all_filters
        if not filters:
            raise ToolkitMissingResourceError("No filters found")
        id_by_display_name = {f"{filter.name} ({filter.external_id})": filter.external_id for filter in filters}
        return tuple(
            questionary.checkbox(
                "Which filters would you like to dump?",
                choices=[Choice(name, value=id_) for name, id_ in id_by_display_name.items()],
            ).ask()
        )

    def _get_filters(self, identifiers: tuple[str, ...]) -> LocationFilterList:
        if not identifiers:
            return self.all_filters
        filters = [f for f in self.all_filters if f.external_id in identifiers]
        if not filters:
            raise ToolkitResourceMissingError(f"Location filters {identifiers} not found", str(identifiers))
        return LocationFilterList(filters)

    def __iter__(self) -> Iterator[tuple[list[Hashable], CogniteResourceList | None, ResourceLoader, None | str]]:
        self.identifier = self.identifier or self._interactive_select()
        filters = self._get_filters(self.identifier)
        yield [], filters, LocationFilterLoader.create_loader(self.client), None


class DumpResourceCommand(ToolkitCommand):
    def dump_to_yamls(
        self,
        finder: ResourceFinder,
        output_dir: Path,
        clean: bool,
        verbose: bool,
    ) -> None:
        is_populated = output_dir.exists() and any(output_dir.iterdir())
        if is_populated and clean:
            safe_rmtree(output_dir)
            output_dir.mkdir()
            self.console(f"Cleaned existing output directory {output_dir!s}.")
        elif is_populated:
            self.warn(MediumSeverityWarning("Output directory is not empty. Use --clean to remove existing files."))
        elif not output_dir.exists():
            output_dir.mkdir(exist_ok=True)

        for identifiers, resources, loader, subfolder in finder:
            if not identifiers and not resources:
                # No resources to dump
                continue
            if resources is None:
                try:
                    resources = loader.retrieve(identifiers)
                except CogniteAPIError as e:
                    raise ResourceRetrievalError(f"Failed to retrieve {humanize_collection(identifiers)}: {e!s}") from e
                if len(resources) == 0:
                    raise ToolkitResourceMissingError(
                        f"Resource(s) {humanize_collection(identifiers)} not found", str(identifiers)
                    )

            finder.update(resources)
            resource_folder = output_dir / loader.folder_name
            if subfolder:
                resource_folder = resource_folder / subfolder
            resource_folder.mkdir(exist_ok=True, parents=True)
            for resource in resources:
                name = loader.as_str(loader.get_id(resource))
                base_filepath = resource_folder / f"{name}.{loader.kind}.yaml"
                if base_filepath.exists():
                    self.warn(FileExistsWarning(base_filepath, "Skipping... Use --clean to remove existing files."))
                    continue
                dumped = loader.dump_resource(resource)
                for filepath, subpart in loader.split_resource(base_filepath, dumped):
                    content = subpart if isinstance(subpart, str) else yaml_safe_dump(subpart)
                    safe_write(filepath, content, encoding="utf-8")
                    if verbose:
                        self.console(f"Dumped {loader.kind} {name} to {filepath!s}")

        print(Panel(f"Dumped {finder.identifier}", title="Success", style="green", expand=False))



from cognite.client.data_classes import (
    Asset,
    CountAggregate,
    DataSet,
)
from questionary import Choice

from cognite_toolkit._cdf_tk.client.testing import monkeypatch_toolkit_client
from cognite_toolkit._cdf_tk.utils.interactive_select import (
    AssetInteractiveSelect,
    EventInteractiveSelect,
    FileMetadataInteractiveSelect,
    TimeSeriesInteractiveSelect,
)
from tests.test_unit.utils import MockQuestionary


class TestInteractiveSelect:
    def test_interactive_select_assets(self, monkeypatch) -> None:
        def select_hierarchy(choices: list[Choice]) -> list[str]:
            assert len(choices) == 2
            return [choices[1].value]

        def select_data_set(choices: list[Choice]) -> list[str]:
            assert len(choices) == 3
            return [choices[2].value]

        answers = ["Hierarchy", select_hierarchy, "Data Set", select_data_set, "Done"]
        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(AssetInteractiveSelect.__module__, monkeypatch, answers),
        ):
            client.assets.list.return_value = [Asset(id=1, external_id="Root1"), Asset(id=2, external_id="Root2")]
            client.assets.aggregate_count.return_value = 100
            client.data_sets.list.return_value = [
                DataSet(id=1, external_id="dataset1"),
                DataSet(id=2, external_id="dataset2"),
                DataSet(id=3, external_id="dataset3"),
            ]

            selector = AssetInteractiveSelect(client)
            selected_hierarchy, selected_dataset = selector.interactive_select_hierarchy_datasets()

        assert selected_hierarchy == ["Root2"]
        assert selected_dataset == ["dataset3"]

    def test_interactive_select_filemetadata(self, monkeypatch) -> None:
        def select_data_set(choices: list[Choice]) -> list[str]:
            assert len(choices) == 3
            return [choices[2].value]

        def select_hierarchy(choices: list[Choice]) -> list[str]:
            assert len(choices) == 2
            return [choices[1].value]

        answers = ["Data Set", select_data_set, "Hierarchy", select_hierarchy, "Done"]
        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(FileMetadataInteractiveSelect.__module__, monkeypatch, answers),
        ):
            client.data_sets.list.return_value = [
                DataSet(id=1, external_id="dataset1", name="Dataset 1"),
                DataSet(id=2, external_id="dataset2", name="Dataset 2"),
                DataSet(id=3, external_id="dataset3", name="Dataset 3"),
            ]
            client.assets.list.return_value = [
                Asset(id=1, external_id="Root1", name="Root 1"),
                Asset(id=2, external_id="Root2", name="Root 2"),
            ]
            client.files.aggregate.return_value = [CountAggregate(100)]
            selector = FileMetadataInteractiveSelect(client)
            selected_hierarchy, selected_dataset = selector.interactive_select_hierarchy_datasets()

        assert selected_hierarchy == ["Root2"]
        assert selected_dataset == ["dataset3"]

    def test_interactive_select_filemetadata_empty_cdf(self, monkeypatch) -> None:
        def select_data_set(choices: list[Choice]) -> list[str]:
            assert len(choices) == 0
            return []

        def select_hierarchy(choices: list[Choice]) -> list[str]:
            assert len(choices) == 0
            return []

        answers = ["Data Set", select_data_set, "Hierarchy", select_hierarchy, "Done"]
        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(FileMetadataInteractiveSelect.__module__, monkeypatch, answers),
        ):
            client.data_sets.list.return_value = []
            client.assets.list.return_value = []
            client.files.aggregate.return_value = [CountAggregate(100)]
            selector = FileMetadataInteractiveSelect(client)
            selected_hierarchy, selected_dataset = selector.interactive_select_hierarchy_datasets()

        assert selected_hierarchy == []
        assert selected_dataset == []

    def test_interactive_select_timeseries(self, monkeypatch) -> None:
        def select_data_set(choices: list[Choice]) -> list[str]:
            assert len(choices) == 3
            return [choices[2].value]

        def select_hierarchy(choices: list[Choice]) -> list[str]:
            assert len(choices) == 2
            return [choices[1].value]

        answers = ["Data Set", select_data_set, "Hierarchy", select_hierarchy, "Done"]
        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(TimeSeriesInteractiveSelect.__module__, monkeypatch, answers),
        ):
            client.data_sets.list.return_value = [
                DataSet(id=1, external_id="dataset1", name="Dataset 1"),
                DataSet(id=2, external_id="dataset2", name="Dataset 2"),
                DataSet(id=3, external_id="dataset3", name="Dataset 3"),
            ]
            client.assets.list.return_value = [
                Asset(id=1, external_id="Root1", name="Root 1"),
                Asset(id=2, external_id="Root2", name="Root 2"),
            ]
            client.time_series.aggregate_count.return_value = 100
            selector = TimeSeriesInteractiveSelect(client)
            selected_hierarchy, selected_dataset = selector.interactive_select_hierarchy_datasets()

        assert selected_hierarchy == ["Root2"]
        assert selected_dataset == ["dataset3"]

    def test_interactive_select_events(self, monkeypatch) -> None:
        def select_data_set(choices: list[Choice]) -> list[str]:
            assert len(choices) == 3
            return [choices[2].value]

        def select_hierarchy(choices: list[Choice]) -> list[str]:
            assert len(choices) == 2
            return [choices[1].value]

        answers = ["Data Set", select_data_set, "Hierarchy", select_hierarchy, "Done"]
        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(EventInteractiveSelect.__module__, monkeypatch, answers),
        ):
            client.data_sets.list.return_value = [
                DataSet(id=1, external_id="dataset1", name="Dataset 1"),
                DataSet(id=2, external_id="dataset2", name="Dataset 2"),
                DataSet(id=3, external_id="dataset3", name="Dataset 3"),
            ]
            client.assets.list.return_value = [
                Asset(id=1, external_id="Root1", name="Root 1"),
                Asset(id=2, external_id="Root2", name="Root 2"),
            ]
            client.events.aggregate_count.return_value = 100
            selector = EventInteractiveSelect(client)
            selected_hierarchy, selected_dataset = selector.interactive_select_hierarchy_datasets()

        assert selected_hierarchy == ["Root2"]
        assert selected_dataset == ["dataset3"]



import os
import shutil
from pathlib import Path

import pytest
from cognite.client import CogniteClient, global_config
from cognite.client.credentials import OAuthClientCredentials
from cognite.client.data_classes.data_modeling import Space, SpaceApply
from dotenv import load_dotenv

from cognite_toolkit._cdf_tk.client import ToolkitClient, ToolkitClientConfig
from cognite_toolkit._cdf_tk.commands import CollectCommand
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables
from tests.constants import REPO_ROOT

THIS_FOLDER = Path(__file__).resolve().parent
TMP_FOLDER = THIS_FOLDER / "tmp"


@pytest.fixture(scope="session")
def toolkit_client_config() -> ToolkitClientConfig:
    load_dotenv(REPO_ROOT / ".env", override=True)
    # Ensure that we do not collect data during tests
    cmd = CollectCommand()
    cmd.execute(action="opt-out")

    cdf_cluster = os.environ["CDF_CLUSTER"]
    credentials = OAuthClientCredentials(
        token_url=os.environ["IDP_TOKEN_URL"],
        client_id=os.environ["IDP_CLIENT_ID"],
        client_secret=os.environ["IDP_CLIENT_SECRET"],
        scopes=[f"https://{cdf_cluster}.cognitedata.com/.default"],
        audience=f"https://{cdf_cluster}.cognitedata.com",
    )
    global_config.disable_pypi_version_check = True
    return ToolkitClientConfig(
        client_name="cdf-toolkit-integration-tests",
        base_url=f"https://{cdf_cluster}.cognitedata.com",
        project=os.environ["CDF_PROJECT"],
        credentials=credentials,
        # We cannot commit auth to WorkflowTrigger and FunctionSchedules.
        is_strict_validation=False,
    )


@pytest.fixture(scope="session")
def cognite_client(toolkit_client_config: ToolkitClientConfig) -> CogniteClient:
    return CogniteClient(toolkit_client_config)


@pytest.fixture(scope="session")
def toolkit_client(toolkit_client_config: ToolkitClientConfig) -> ToolkitClient:
    return ToolkitClient(toolkit_client_config)


@pytest.fixture()
def max_two_workers():
    old = global_config.max_workers
    global_config.max_workers = 2
    yield
    global_config.max_workers = old


@pytest.fixture(scope="session")
def toolkit_client_with_pending_ids(toolkit_client_config: ToolkitClientConfig) -> ToolkitClient:
    """Returns a ToolkitClient configured to enable pending IDs."""
    return ToolkitClient(toolkit_client_config, enable_set_pending_ids=True)


@pytest.fixture(scope="session")
def env_vars(toolkit_client: ToolkitClient) -> EnvironmentVariables:
    env_vars = EnvironmentVariables.create_from_environment()
    # Ensure we use the client above that has CLIENT NAME set to the test name
    env_vars._client = toolkit_client
    return env_vars


@pytest.fixture(scope="session")
def toolkit_space(cognite_client: CogniteClient) -> Space:
    return cognite_client.data_modeling.spaces.apply(SpaceApply(space="toolkit_test_space"))


@pytest.fixture
def build_dir() -> Path:
    pidid = os.getpid()
    build_path = TMP_FOLDER / f"build-{pidid}"
    build_path.mkdir(exist_ok=True, parents=True)
    yield build_path
    shutil.rmtree(build_path, ignore_errors=True)


@pytest.fixture(scope="session")
def dev_cluster_client() -> ToolkitClient | None:
    """Returns a ToolkitClient configured for the development cluster."""
    dev_cluster_env = REPO_ROOT / "dev-cluster.env"
    if not dev_cluster_env.exists():
        pytest.skip("dev-cluster.env file not found, skipping tests that require dev cluster client.")
        return None
    env_content = dev_cluster_env.read_text(encoding="utf-8")
    env_vars = dict(
        line.strip().split("=")
        for line in env_content.splitlines()
        if line.strip() and not line.startswith("#") and "=" in line
    )
    cdf_cluster = env_vars["CDF_CLUSTER"]
    credentials = OAuthClientCredentials(
        token_url=env_vars["IDP_TOKEN_URL"],
        client_id=env_vars["IDP_CLIENT_ID"],
        client_secret=env_vars["IDP_CLIENT_SECRET"],
        scopes=[f"https://{cdf_cluster}.cognitedata.com/.default"],
        audience=f"https://{cdf_cluster}.cognitedata.com",
    )
    config = ToolkitClientConfig(
        client_name="cdf-toolkit-integration-tests",
        base_url=f"https://{cdf_cluster}.cognitedata.com",
        project=env_vars["CDF_PROJECT"],
        credentials=credentials,
        is_strict_validation=False,
    )
    return ToolkitClient(config, enable_set_pending_ids=True)



"""Gauge reader handler."""

import io
import json
import logging
import sys
from typing import Any, Optional

import numpy as np
from bosdyn.client.math_helpers import Quat, SE3Pose
from cognite.client import CogniteClient, global_config
from cognite.client.data_classes import FileMetadata, FileMetadataList, FileMetadataUpdate, LabelFilter
from cognite.client.exceptions import CogniteDuplicatedError
from cognite_threesixty_images import CogniteThreeSixtyImageExtractor, VectorXYZ
from PIL import Image
from scipy.spatial.transform import Rotation

logging.basicConfig(stream=sys.stdout, level=logging.INFO, format=logging.BASIC_FORMAT)
logger = logging.getLogger(__name__)
CUBEMAP_RESOLUTION = 1024
global_config.disable_gzip = True


def get_map_transform_from_map(client: CogniteClient, map_external_id: str) -> Optional[dict]:
    """Get the transform of a map in the current project based on mission. We asume the frame has the root frame as parent."""
    project = client._config.project
    map_request = {"items": [{"externalId": map_external_id}]}

    res = client.post(url=f"/api/v1/projects/{project}/robotics/maps/byids", json=map_request).json()
    for map in res["items"]:
        frame_external_id = map.get("frameExternalId")
    assert frame_external_id, f"Map {map_external_id} does not have a frame associated with it."
    # Create mapping
    frame_request = {"items": [{"externalId": frame_external_id}]}

    res = client.post(
        url=f"/api/v1/projects/{project}/robotics/frames/byids",
        json=frame_request,
    ).json()

    print(f"going to get frame {frame_external_id} from map {map_external_id}")

    for item in res["items"]:
        if item.get("transform"):
            print(f"found translation {item.get('transform').get('orientation').get('w')}")
            return SE3Pose(
                x=item.get("transform").get("translation").get("x"),
                y=item.get("transform").get("translation").get("y"),
                z=item.get("transform").get("translation").get("z"),
                rot=Quat(
                    w=item.get("transform").get("orientation").get("w"),
                    x=item.get("transform").get("orientation").get("x"),
                    y=item.get("transform").get("orientation").get("y"),
                    z=item.get("transform").get("orientation").get("z"),
                ),
            )
    return None


def convert_metadata_to_se3_pose(external_id: str, metadata: dict[str, Any]) -> Optional[SE3Pose]:
    """Convert metadata to SE3Pose."""
    x = metadata.get("waypoint_tform_body_x")
    y = metadata.get("waypoint_tform_body_x")
    z = metadata.get("waypoint_tform_body_x")
    qx = metadata.get("waypoint_tform_body_qx")
    qy = metadata.get("waypoint_tform_body_qy")
    qz = metadata.get("waypoint_tform_body_qz")
    qw = metadata.get("waypoint_tform_body_qw")
    if not x or not y or not z or not qx or not qy or not qz or not qw:
        logger.error(
            f"Failed to process file {external_id}. Missing metadata field. Required metadata fields \
            are waypoint_tform_body_x, waypoint_tform_body_x, waypoint_tform_body_x, waypoint_tform_body_qx, \
            waypoint_tform_body_qy, waypoint_tform_body_qz, waypoint_tform_body_qw. File metadata keys: {metadata.keys()}"
        )
        return None
    waypoint_tform_body = SE3Pose(
        x=float(x), y=float(y), z=float(z), rot=Quat(w=float(qw), x=float(qx), y=float(qy), z=float(qz))
    )
    return waypoint_tform_body


def get_waypoint_and_pose(
    waypoint_id: str, client: CogniteClient
) -> tuple[Optional[dict[str, Any]], Optional[SE3Pose]]:
    """Get a waypoint from the robotics api and return the waypoint and pose."""
    get_waypoints_body = {"items": [{"externalId": waypoint_id}]}
    get_waypoints_response = client.post(
        f"/api/v1/projects/{client.config.project}/robotics/waypoints/byids", json=get_waypoints_body
    )
    waypoint = json.loads(get_waypoints_response.content)["items"]
    if len(waypoint) != 1:
        logger.error(
            f"Failed to process threesixty image. Did not get exactly 1 waypoint with external id {waypoint_id}."
        )
        return None, None
    waypoint_pos = waypoint[0].get("position")
    waypoint_ori = waypoint[0].get("orientation")
    ko_tform_waypoint = SE3Pose(
        x=waypoint_pos["x"],
        y=waypoint_pos["y"],
        z=waypoint_pos["z"],
        rot=Quat(w=waypoint_ori["w"], x=waypoint_ori["x"], y=waypoint_ori["y"], z=waypoint_ori["z"]),
    )
    return waypoint[0], ko_tform_waypoint


def create_and_upload_360_files(
    cognite_client: CogniteClient,
    cognite_threesixty_image_extractor: CogniteThreeSixtyImageExtractor,
    robot_pose: SE3Pose,
    image: Image,
    waypoint: dict[str, Any],
    timestamp: int,
):
    """Create Cogntie three sixty images."""
    rot = Rotation.from_quat(
        [
            robot_pose.rotation.x,
            robot_pose.rotation.y,
            robot_pose.rotation.z,
            robot_pose.rotation.w,
        ]
    )
    rot_vec = rot.as_rotvec()
    rot_angle = np.linalg.norm(rot_vec)
    rot_vec = rot_vec / rot_angle

    event, files = cognite_threesixty_image_extractor.create_threesixty_image(
        content=np.array(image),
        site_id=waypoint.get("mapExternalId", "default_location"),
        site_name=waypoint.get("mapExternalId", "default_location"),
        station_number=waypoint.get("externalId"),  # TODO: make station id more readable
        rotation_angle=str(rot_angle),
        rotation_axis=VectorXYZ(float(rot_vec[0]), float(rot_vec[1]), float(rot_vec[2])),
        rotation_angle_unit="rad",
        translation=VectorXYZ(robot_pose.x, robot_pose.y, robot_pose.z),
        translation_unit="m",
        translation_offset_mm=VectorXYZ(0, 0, 0),
        timestamp=timestamp,
    )
    logger.info("Created event and files")

    # Upload event
    try:
        cognite_client.events.create([event])
    except CogniteDuplicatedError:
        logger.warning(f"Event with external id {event.external_id} already exists. will update event.")
        cognite_client.events.update([event])

    logger.info(f"Created event in CDF with external id {event.external_id}")
    # Upload files
    for file in files:
        cognite_client.files.upload_bytes(
            content=file.content,
            name=file.file_metadata.name,
            data_set_id=file.file_metadata.data_set_id,
            external_id=file.file_metadata.external_id,
            source=file.file_metadata.source,
            mime_type=file.file_metadata.mime_type,
            metadata=file.file_metadata.metadata,
            asset_ids=file.file_metadata.asset_ids,
            labels=file.file_metadata.labels,
            overwrite=True,
        )
        logger.info(f"Created file in CDF with external id {file.file_metadata.external_id}")
    logger.info("Completed uploading 360 image to CDF.")


def process_threesixty_files(files: FileMetadataList, client: CogniteClient, data_set_id: int):
    """Process three sixty images."""
    cognite_threesixty_image_extractor = CogniteThreeSixtyImageExtractor(data_set_id=data_set_id)

    file: FileMetadata
    for file in files:
        try:
            # Updating the file immediately so that the same file will not be processed again
            client.files.update(FileMetadataUpdate(id=file.id).labels.remove("threesixty"))

            logger.info(f"Processing file with external id {file.external_id}, id {file.id}")
            if not file.uploaded:
                logger.error(f"file not upload not completed {file.external_id}")
                continue

            # Get waypoint id from metadata
            waypoint_id = file.metadata.get("waypoint_id")
            print(f"Waypoint id: {waypoint_id}")
            if not waypoint_id:
                client.files.update(FileMetadataUpdate(id=file.id).labels.remove("threesixty"))
                logger.error(f"Failed to process file {file.external_id}. No waypoint id in the metadata.")
                continue

            # Calculate robot pose
            waypoint_tform_body = convert_metadata_to_se3_pose(file.external_id, file.metadata)
            if waypoint_tform_body is None:
                continue
            waypoint, ko_tform_waypoint = get_waypoint_and_pose(waypoint_id=waypoint_id, client=client)
            if waypoint_tform_body is None or waypoint is None:
                continue
            robot_pose = ko_tform_waypoint * waypoint_tform_body

            print(f"waypoint : {waypoint}")

            site_alignement = get_map_transform_from_map(
                client=client, map_external_id=str(waypoint.get("mapExternalId"))
            )

            site_pose = site_alignement * robot_pose

            # Download file
            image_data = client.files.download_bytes(id=file.id)
            try:
                image = Image.open(io.BytesIO(image_data))
            except Exception:
                logger.error(f"This CDF File does not seem to be an image. File ID: {file.id}")
                continue

            # If images don't have a timestamp metadata field, default to the created time of the image file
            image_timestamp = file.metadata.get("timestamp")
            if image_timestamp is None:
                image_timestamp = file.created_time

            # Create and upload 360 files
            create_and_upload_360_files(
                cognite_client=client,
                cognite_threesixty_image_extractor=cognite_threesixty_image_extractor,
                robot_pose=site_pose,
                image=image,
                waypoint=waypoint,
                timestamp=image_timestamp,
            )
        except Exception as e:
            client.files.update(FileMetadataUpdate(id=file.id).labels.add("threesixty"))
            logger.error(f"Failed to process file {file}. Error: {e}")


def handle(data, client):
    """Three sixty image handle. ."""
    logger.info("Start three sixty processing.")

    # Check that input contains data_set_id
    if "data_set_external_id" not in data.keys():
        raise RuntimeError("Data should contain all keys: data_set_id")
    data_set_id = client.data_sets.retrieve(external_id=data["data_set_external_id"]).id

    # Get all 360 images in the data set id with the label "threesixty"
    # Changed default limit to 1 to process 1 file at a time as the upload_queue in cognite_threesixty_image uploads the whole queue
    # on every file but after the first cycle fails as the first event is duplicated and the whole queue upload fails
    files: FileMetadataList = client.files.list(
        labels=LabelFilter(contains_all=["threesixty"]),
        limit=25,
        data_set_ids=[data_set_id],
        uploaded=True,
    )
    logger.info(f"Processing {len(files)} threesixty files.")
    # Process three sixty files
    process_threesixty_files(files=files, client=client, data_set_id=data_set_id)
    return {}



from __future__ import annotations

from pathlib import Path
from unittest.mock import MagicMock

import pytest

from cognite_toolkit._cdf_tk.builders import get_loader
from cognite_toolkit._cdf_tk.loaders import FileLoader, RawDatabaseLoader, RawTableLoader, ResourceLoader


@pytest.mark.parametrize(
    "content, expected_loader_cls",
    [
        ("dbName: my_database\n", RawDatabaseLoader),
        ("dbName: my_database\ntableName: my_table\n", RawTableLoader),
    ],
)
def test_get_loader_raw_loaders(content: str, expected_loader_cls: type[ResourceLoader]) -> None:
    filepath = MagicMock(spec=Path)
    filepath.name = "filelocation.yaml"
    filepath.stem = "filelocation"
    filepath.suffix = ".yaml"
    filepath.read_text.return_value = content

    loader, warn = get_loader(filepath, "raw", force_pattern=True)

    assert warn is None
    assert loader is expected_loader_cls


def test_get_loader_file() -> None:
    loader_cls, warning = get_loader(Path("SHOP_model_borgund.File.yaml"), "files", force_pattern=True)

    assert warning is None
    assert loader_cls is FileLoader



from cognite_toolkit import _version
from cognite_toolkit._cdf_tk.cdf_toml import CDFToml
from tests.constants import REPO_ROOT


class TestCDFToml:
    def test_load_repo_root_config(self) -> None:
        config = CDFToml.load(REPO_ROOT)

        assert config.modules.version == _version.__version__



from __future__ import annotations

from pathlib import Path
from unittest.mock import patch

from pytest import raises

from cognite_toolkit._cdf_tk.constants import BUILTIN_MODULES_PATH
from cognite_toolkit._cdf_tk.data_classes._module_toml import ModuleToml
from cognite_toolkit._cdf_tk.data_classes._packages import Packages


class TestPackages:
    def test_load(self) -> None:
        packages = Packages.load(BUILTIN_MODULES_PATH)
        assert packages is not None
        assert len(packages) >= 5
        assert "infield" in packages
        infield = packages["infield"]
        assert len(infield.modules) > 0
        assert "cdf_infield_common" in infield.module_names
        assert "cdf_infield_location" in infield.module_names

    def test_fail_on_invalid_tag(self) -> None:
        def mock_toml_load(path: Path) -> ModuleToml:
            return ModuleToml("description", frozenset(["invalid_tag", "tag2"]))

        with patch("cognite_toolkit._cdf_tk.data_classes._module_toml.ModuleToml.load", side_effect=mock_toml_load):
            # Test the overridden behavior
            with raises(ValueError):
                Packages.load(BUILTIN_MODULES_PATH)



"""This is an internal subpackage used to check the parameters of a class."""

from .constants import ANY_INT, ANY_STR, ANYTHING
from .data_classes import ParameterSet, ParameterSpec, ParameterSpecSet, ParameterValue
from .functions import read_parameter_from_init_type_hints, read_parameters_from_dict

__all__ = [
    "ANYTHING",
    "ANY_INT",
    "ANY_STR",
    "ParameterSet",
    "ParameterSpec",
    "ParameterSpecSet",
    "ParameterSpecSet",
    "ParameterValue",
    "read_parameter_from_init_type_hints",
    "read_parameters_from_dict",
]



from pathlib import Path
from typing import Annotated, Any

import typer

from cognite_toolkit._cdf_tk.commands import MigrateTimeseriesCommand
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables


class MigrateApp(typer.Typer):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.callback(invoke_without_command=True)(self.main)
        self.command("timeseries")(self.timeseries)

    def main(self, ctx: typer.Context) -> None:
        """Migrate resources from Asset-Centric to data modeling in CDF."""
        if ctx.invoked_subcommand is None:
            print("Use [bold yellow]cdf migrate --help[/] for more information.")

    @staticmethod
    def timeseries(
        ctx: typer.Context,
        mapping_file: Annotated[
            Path,
            typer.Option(
                "--mapping-file",
                "-m",
                help="Path to the mapping file that contains the mapping from TimeSeries to CogniteTimeSeries. "
                "This file is expected to have the following columns: [id/externalId, dataSetId, space, externalId]."
                "The dataSetId is optional, and can be skipped. If it is set, it is used to check the access to the dataset.",
            ),
        ],
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-d",
                help="If set, the migration will not be executed, but only a report of what would be done is printed.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """Migrate TimeSeries to CogniteTimeSeries."""

        client = EnvironmentVariables.create_from_environment().get_client(enable_set_pending_ids=True)
        cmd = MigrateTimeseriesCommand()
        cmd.run(
            lambda: cmd.migrate_timeseries(
                client,
                mapping_file=mapping_file,
                dry_run=dry_run,
                verbose=verbose,
            )
        )



from __future__ import annotations

import contextlib
import re
from collections import defaultdict
from collections.abc import Hashable, Iterable, Sequence
from pathlib import Path
from typing import Any, Literal, cast

import yaml
from cognite.client.exceptions import CogniteAPIError
from rich import print
from rich.panel import Panel
from rich.progress import track

from cognite_toolkit._cdf_tk.builders import Builder, create_builder
from cognite_toolkit._cdf_tk.cdf_toml import CDFToml
from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.raw import RawDatabase
from cognite_toolkit._cdf_tk.commands._base import ToolkitCommand
from cognite_toolkit._cdf_tk.constants import (
    _RUNNING_IN_BROWSER,
    BUILD_FOLDER_ENCODING,
    DEFAULT_ENV,
    DEV_ONLY_MODULES,
    HINT_LEAD_TEXT,
    ROOT_MODULES,
    TEMPLATE_VARS_FILE_SUFFIXES,
    URL,
    YAML_SUFFIX,
)
from cognite_toolkit._cdf_tk.data_classes import (
    BuildConfigYAML,
    BuildDestinationFile,
    BuildSourceFile,
    BuildVariables,
    BuiltModule,
    BuiltModuleList,
    BuiltResource,
    BuiltResourceList,
    ModuleDirectories,
    ModuleLocation,
    SourceLocation,
    SourceLocationEager,
    SourceLocationLazy,
)
from cognite_toolkit._cdf_tk.exceptions import (
    ToolkitDuplicatedModuleError,
    ToolkitEnvError,
    ToolkitError,
    ToolkitMissingModuleError,
    ToolkitYAMLFormatError,
)
from cognite_toolkit._cdf_tk.hints import Hint, ModuleDefinition, verify_module_directory
from cognite_toolkit._cdf_tk.loaders import (
    ContainerLoader,
    DataLoader,
    DataModelLoader,
    DataSetsLoader,
    ExtractionPipelineConfigLoader,
    FileLoader,
    LocationFilterLoader,
    NodeLoader,
    RawDatabaseLoader,
    RawTableLoader,
    ResourceLoader,
    SpaceLoader,
    TransformationLoader,
    ViewLoader,
)
from cognite_toolkit._cdf_tk.tk_warnings import (
    DuplicatedItemWarning,
    FileReadWarning,
    LowSeverityWarning,
    MediumSeverityWarning,
    MissingDependencyWarning,
    UnresolvedVariableWarning,
    WarningList,
)
from cognite_toolkit._cdf_tk.tk_warnings.fileread import MissingRequiredIdentifierWarning
from cognite_toolkit._cdf_tk.utils import (
    calculate_hash,
    humanize_collection,
    quote_int_value_by_key_in_yaml,
    read_yaml_content,
    safe_read,
    safe_write,
    stringify_value_by_key_in_yaml,
)
from cognite_toolkit._cdf_tk.utils.file import safe_rmtree
from cognite_toolkit._cdf_tk.utils.modules import parse_user_selected_modules
from cognite_toolkit._cdf_tk.validation import (
    validate_data_set_is_set,
    validate_modules_variables,
    validate_resource_yaml,
    validate_resource_yaml_pydantic,
)
from cognite_toolkit._version import __version__


class BuildCommand(ToolkitCommand):
    def __init__(
        self,
        print_warning: bool = True,
        skip_tracking: bool = False,
        silent: bool = False,
    ) -> None:
        super().__init__(print_warning, skip_tracking, silent)
        self.existing_resources_by_loader: dict[type[ResourceLoader], set[Hashable]] = defaultdict(set)
        self.instantiated_loaders: dict[type[ResourceLoader], ResourceLoader] = {}

        # Built State
        self._module_names_by_variable_key: dict[str, list[str]] = defaultdict(list)
        self._builder_by_resource_folder: dict[str, Builder] = {}
        self._ids_by_resource_type: dict[type[ResourceLoader], dict[Hashable, SourceLocation]] = defaultdict(dict)
        self._dependencies_by_required: dict[tuple[type[ResourceLoader], Hashable], list[tuple[Hashable, Path]]] = (
            defaultdict(list)
        )
        self._has_built = False
        self._printed_variable_tree_structure_hint = False

    def execute(
        self,
        verbose: bool,
        organization_dir: Path,
        build_dir: Path,
        selected: list[str | Path] | None,
        build_env_name: str | None,
        no_clean: bool,
        client: ToolkitClient | None = None,
        on_error: Literal["continue", "raise"] = "continue",
    ) -> BuiltModuleList:
        if organization_dir in {Path("."), Path("./")}:
            organization_dir = Path.cwd()
        verify_module_directory(organization_dir, build_env_name)

        cdf_toml = CDFToml.load()

        if (organization_dir / BuildConfigYAML.get_filename(build_env_name or DEFAULT_ENV)).exists():
            config = BuildConfigYAML.load_from_directory(organization_dir, build_env_name or DEFAULT_ENV)
        else:
            # Loads the default environment
            config = BuildConfigYAML.load_default(organization_dir)

        if selected:
            config.environment.selected = parse_user_selected_modules(selected, organization_dir)

        directory_name = "current directory" if organization_dir == Path(".") else f"project '{organization_dir!s}'"
        root_modules = [
            module_dir for root_module in ROOT_MODULES if (module_dir := organization_dir / root_module).exists()
        ]
        module_locations = "\n".join(f"  - Module directory '{root_module!s}'" for root_module in root_modules)
        print(
            Panel(
                f"Building {directory_name}:\n  - Toolkit Version '{__version__!s}'\n"
                f"  - Environment name {build_env_name!r}, validation-type {config.environment.validation_type!r}.\n"
                f"  - Config '{config.filepath!s}'"
                f"\n{module_locations}",
                expand=False,
            )
        )

        config.set_environment_variables()

        return self.build_config(
            build_dir=build_dir,
            organization_dir=organization_dir,
            config=config,
            packages=cdf_toml.modules.packages,
            clean=not no_clean,
            verbose=verbose,
            client=client,
            on_error=on_error,
        )

    def build_config(
        self,
        build_dir: Path,
        organization_dir: Path,
        config: BuildConfigYAML,
        packages: dict[str, list[str]],
        clean: bool = False,
        verbose: bool = False,
        client: ToolkitClient | None = None,
        progress_bar: bool = False,
        on_error: Literal["continue", "raise"] = "continue",
    ) -> BuiltModuleList:
        is_populated = build_dir.exists() and any(build_dir.iterdir())
        if is_populated and clean:
            safe_rmtree(build_dir)
            build_dir.mkdir()
            if not _RUNNING_IN_BROWSER:
                self.console(f"Cleaned existing build directory {build_dir!s}.")
        elif is_populated and not _RUNNING_IN_BROWSER:
            self.warn(
                LowSeverityWarning("Build directory is not empty. Run without --no-clean to remove existing files.")
            )
        elif build_dir.exists() and not _RUNNING_IN_BROWSER:
            self.console("Build directory does already exist and is empty. No need to create it.")
        else:
            build_dir.mkdir(exist_ok=True)

        if issue := config.validate_environment():
            self.warn(issue)

        user_selected_modules = config.environment.get_selected_modules(packages)
        modules = ModuleDirectories.load(organization_dir, user_selected_modules)
        self._validate_modules(modules, config, packages, user_selected_modules, organization_dir)

        if verbose:
            self.console("Selected packages:")
            selected_packages = [package for package in packages if package in config.environment.selected]
            if len(selected_packages) == 0:
                self.console("    None", prefix="")
            for package in selected_packages:
                self.console(f"    {package}", prefix="")
            self.console("Selected modules:")
            for module in [module.name for module in modules.selected]:
                self.console(f"    {module}", prefix="")

        variables = BuildVariables.load_raw(
            config.variables, modules.available_paths, modules.selected.available_paths, config.filepath
        )
        warnings = validate_modules_variables(variables.selected, config.filepath)
        if warnings:
            self.console(
                f"Found the following warnings in config.{config.environment.name}.yaml:",
                prefix="[bold red]Warning:[/]",
            )
            for warning in warnings:
                if self.print_warning:
                    print(f"    {warning.get_message()}")

        # Setup state before building modules
        self._module_names_by_variable_key.clear()
        self._builder_by_resource_folder.clear()
        for variable in variables:
            for module_location in modules:
                if variable.location in module_location.relative_path.parts:
                    self._module_names_by_variable_key[variable.key].append(module_location.name)
        if self._has_built:
            # Todo: Reset of state??
            raise RuntimeError("In the build command, the `build_config` method should only be called once.")
        else:
            self._has_built = True

        built_modules = self.build_modules(modules.selected, build_dir, variables, verbose, progress_bar, on_error)

        self._check_missing_dependencies(organization_dir, client)

        build_environment = config.create_build_environment(built_modules, modules.selected)
        build_environment.dump_to_file(build_dir)
        if not _RUNNING_IN_BROWSER:
            self.console(f"Build complete. Files are located in {build_dir!s}/")
        return built_modules

    def build_modules(
        self,
        modules: ModuleDirectories,
        build_dir: Path,
        variables: BuildVariables,
        verbose: bool = False,
        progress_bar: bool = False,
        on_error: Literal["continue", "raise"] = "continue",
    ) -> BuiltModuleList:
        build = BuiltModuleList()
        warning_count = len(self.warning_list)
        if progress_bar:
            modules_iter = cast(
                Iterable[ModuleLocation], track(modules, description="Building modules", transient=True)
            )
        else:
            modules_iter = modules
        for module in modules_iter:
            if verbose:
                self.console(f"Processing module {module.name}")
            module_variable_sets = variables.get_module_variables(module)
            last_identifiers: set[tuple[Hashable, Path]] = set()
            for iteration, module_variables in enumerate(module_variable_sets, 1):
                try:
                    built_module_resources = self._build_module_resources(module, build_dir, module_variables, verbose)
                except ToolkitError as err:
                    if on_error == "raise":
                        raise

                    suffix = "" if len(module_variable_sets) == 1 else f" ({iteration} of {len(module_variable_sets)})"

                    print(f"  [bold red]Failed Building:([/][red]: {module.name}{suffix}")
                    print(f"  [bold red]ERROR ([/][red]{type(err).__name__}[/][bold red]):[/] {err}")
                    built_status = type(err).__name__
                    built_module_resources = {}
                else:
                    built_status = "Success"

                # Check for duplicates
                identifiers = {
                    (resource.identifier, resource.source.path)
                    for resources in built_module_resources.values()
                    for resource in resources
                }
                if duplicates := (identifiers & last_identifiers):
                    duplicate_warnings = WarningList[FileReadWarning]()
                    for identifier, path in duplicates:
                        duplicate_warnings.append(DuplicatedItemWarning(path, identifier, path))
                    self.warning_list.extend(duplicate_warnings)
                    print(str(duplicate_warnings))
                    print(
                        f"    {HINT_LEAD_TEXT}This is likely due to missing variable in the "
                        f"identifier when using the module {module.name!r} as a template."
                    )
                last_identifiers = identifiers

                module_warnings = len(self.warning_list) - warning_count
                warning_count = len(self.warning_list)

                name = module.name if len(module_variable_sets) == 1 else f"{module.name} (iteration {iteration})"
                built_module = BuiltModule(
                    name=name,
                    location=SourceLocationLazy(
                        path=module.relative_path,
                        absolute_path=module.dir,
                    ),
                    build_variables=module_variables,
                    resources=built_module_resources,
                    warning_count=module_warnings,
                    status=built_status,
                    iteration=iteration,
                )
                build.append(built_module)
                self.tracker.track_module_build(built_module)
        return build

    def _build_module_resources(
        self,
        module: ModuleLocation,
        build_dir: Path,
        module_variables: BuildVariables,
        verbose: bool,
    ) -> dict[str, BuiltResourceList]:
        build_resources_by_folder: dict[str, BuiltResourceList] = defaultdict(BuiltResourceList)
        if not_resource_directory := module.not_resource_directories:
            self.warn(
                LowSeverityWarning(
                    f"Module {module.dir.name!r} has non-resource directories: {sorted(not_resource_directory)}. {ModuleDefinition.short()}"
                )
            )

        for resource_name, resource_files in module.source_paths_by_resource_folder.items():
            source_files = self._replace_variables(resource_files, module_variables, resource_name, module.dir, verbose)

            builder = self._get_builder(build_dir, resource_name)

            built_resources = BuiltResourceList[Hashable]()
            for destination in builder.build(source_files, module):
                if not isinstance(destination, BuildDestinationFile):
                    for warning in destination:
                        self.warn(warning)
                    continue
                if destination.loader is FileLoader:
                    # This is a content file that we should not copy to the build directory.
                    continue

                safe_write(destination.path, destination.content, encoding=BUILD_FOLDER_ENCODING)
                if issubclass(destination.loader, DataLoader):
                    continue

                file_warnings, identifiers_kind_pairs = self.check_built_resource(
                    destination.loaded,
                    destination.loader,
                    destination.source,
                )
                file_warnings.extend(destination.warnings)

                if file_warnings:
                    self.warning_list.extend(file_warnings)
                    # Here we do not use the self.warn method as we want to print the warnings as a group.
                    if self.print_warning:
                        print(str(file_warnings))

                built_source = BuiltResourceList(
                    [
                        BuiltResource(
                            identifier,
                            destination.source,
                            kind,
                            destination.path,
                            extra_sources=destination.extra_sources,
                        )
                        for identifier, kind in identifiers_kind_pairs
                    ]
                )
                built_resources.extend(built_source)

            builder.validate_directory(built_resources, module)

            build_resources_by_folder[resource_name].extend(built_resources)

        return build_resources_by_folder

    def _get_builder(self, build_dir: Path, resource_name: str) -> Builder:
        if resource_name not in self._builder_by_resource_folder:
            self._builder_by_resource_folder[resource_name] = create_builder(resource_name, build_dir, self.warn)
        builder = self._builder_by_resource_folder[resource_name]
        return builder

    def _validate_modules(
        self,
        modules: ModuleDirectories,
        config: BuildConfigYAML,
        packages: dict[str, list[str]],
        selected_modules: set[str | Path],
        organization_dir: Path,
    ) -> None:
        # Validations: Ambiguous selection.
        selected_names = {s for s in config.environment.selected if isinstance(s, str)}
        if duplicate_modules := {
            module_name: paths
            for module_name, paths in modules.as_path_by_name().items()
            if len(paths) > 1 and module_name in selected_names
        }:
            # If the user has selected a module by name, and there are multiple modules with that name, raise an error.
            # Note, if the user uses a path to select a module, this error will not be raised.
            raise ToolkitDuplicatedModuleError(
                f"Ambiguous module selected in config.{config.environment.name}.yaml:", duplicate_modules
            )
        # Package Referenced Modules Exists
        for package, package_modules in packages.items():
            if package not in selected_names:
                # We do not check packages that are not selected.
                # Typically, the user will delete the modules that are irrelevant for them;
                # thus we only check the selected packages.
                continue
            if missing_packages := set(package_modules) - modules.available_names:
                ToolkitMissingModuleError(
                    f"Package {package} defined in {CDFToml.file_name!s} is referring "
                    f"the following missing modules {missing_packages}."
                )

        # Selected modules does not exists
        if missing_modules := set(selected_modules) - modules.available:
            hint = ModuleDefinition.long(missing_modules, organization_dir)
            raise ToolkitMissingModuleError(
                f"The following selected modules are missing, please check path: {missing_modules}.\n{hint}"
            )

        # Nothing is Selected
        if not modules.selected:
            raise ToolkitEnvError(
                f"No selected modules specified in {config.filepath!s}, have you configured "
                f"the environment ({config.environment.name})?"
            )

        dev_modules = modules.available_names & DEV_ONLY_MODULES
        if dev_modules and config.environment.validation_type != "dev":
            self.warn(
                MediumSeverityWarning(
                    "The following modules should [bold]only[/bold] be used a in CDF Projects designated as dev (development): "
                    f"{humanize_collection(dev_modules)!r}",
                )
            )

    def _replace_variables(
        self,
        resource_files: Sequence[Path],
        variables: BuildVariables,
        resource_name: str,
        module_dir: Path,
        verbose: bool,
    ) -> list[BuildSourceFile]:
        source_files: list[BuildSourceFile] = []

        for source_path in resource_files:
            if source_path.suffix.lower() not in TEMPLATE_VARS_FILE_SUFFIXES:
                continue

            if verbose:
                self.console(f"Processing file {source_path.name}...")

            content = safe_read(source_path)
            # We cannot use the content as the basis for hash as this have been encoded.
            # Instead, we use the source path, which will hash the bytes of the file directly,
            # which is what we do in the deploy step to verify that the source file has not changed.
            source = SourceLocationEager(source_path, calculate_hash(source_path, shorten=True))

            content = variables.replace(content, source_path.suffix)

            replace_warnings = self._check_variables_replaced(content, module_dir, source_path)

            if source_path.suffix not in YAML_SUFFIX:
                source_files.append(BuildSourceFile(source, content, None))
                continue

            if resource_name in {
                TransformationLoader.folder_name,
                DataModelLoader.folder_name,
                LocationFilterLoader.folder_name,
            }:
                # Ensure that all keys that are version gets read as strings.
                # This is required by DataModels, Views, and Transformations that reference DataModels and Views.
                content = quote_int_value_by_key_in_yaml(content, key="version")

            if resource_name in ExtractionPipelineConfigLoader.folder_name:
                # Ensure that the config variables are stings.
                # This is required by ExtractionPipelineConfig
                content = stringify_value_by_key_in_yaml(content, key="config")
            try:
                loaded = read_yaml_content(content)
            except yaml.YAMLError as e:
                message = (
                    f"YAML validation error for {source_path.as_posix()!r} after substituting config variables:\n{e}"
                )
                if unresolved_variables := [
                    w.variable.removesuffix("}}").removeprefix("{{").strip()
                    for w in replace_warnings
                    if isinstance(w, UnresolvedVariableWarning)
                ]:
                    variable_str = humanize_collection(set(unresolved_variables))
                    source_str = variables.source_path.as_posix() if variables.source_path else "config.[ENV].yaml"
                    suffix = "s" if len(unresolved_variables) > 1 else ""
                    message += f"\n{HINT_LEAD_TEXT}Add the following variable{suffix} to the {source_str!r} file: {variable_str!r}."
                    message += (
                        f"\nRead more about build variables: {Hint.link(URL.build_variables, URL.build_variables)}."
                    )

                raise ToolkitYAMLFormatError(message)

            source_files.append(BuildSourceFile(source, content, loaded))

        return source_files

    def _check_variables_replaced(self, content: str, module: Path, source_path: Path) -> WarningList[FileReadWarning]:
        all_unmatched = re.findall(pattern=r"\{\{.*?\}\}", string=content)
        warning_list = WarningList[FileReadWarning]()
        for unmatched in all_unmatched:
            warning_list.append(UnresolvedVariableWarning(source_path, unmatched))
            variable = unmatched[2:-2]
            if module_names := self._module_names_by_variable_key.get(variable):
                module_str = (
                    f"{module_names[0]!r}"
                    if len(module_names) == 1
                    else (", ".join(module_names[:-1]) + f" or {module_names[-1]}")
                )
                if not self._printed_variable_tree_structure_hint:
                    self._printed_variable_tree_structure_hint = True
                    self.console(
                        f"The variables in 'config.[ENV].yaml' need to be organised in a tree structure following"
                        f"\n    the folder structure of the modules, but can also be moved up the config hierarchy to be shared between modules."
                        f"\n    The variable {variable!r} is defined in the variable section{'s' if len(module_names) > 1 else ''} {module_str}."
                        f"\n    Check that {'these paths reflect' if len(module_names) > 1 else 'this path reflects'} "
                        f"the location of {module.as_posix()}.",
                        prefix="    [bold green]Hint:[/] ",
                    )
        self.warning_list.extend(warning_list)
        if self.print_warning and warning_list:
            print(str(warning_list))
        return warning_list

    def _check_missing_dependencies(self, project_config_dir: Path, client: ToolkitClient | None = None) -> None:
        existing = {(resource_cls, id_) for resource_cls, ids in self._ids_by_resource_type.items() for id_ in ids}
        missing_dependencies = set(self._dependencies_by_required.keys()) - existing
        for loader_cls, id_ in missing_dependencies:
            if self._is_system_resource(loader_cls, id_):
                continue
            elif loader_cls is DataSetsLoader and id_ == "":
                # Special case used by the location filter to indicate filter out all classical resources.
                continue
            elif client and self._check_resource_exists_in_cdf(client, loader_cls, id_):
                continue
            elif loader_cls.resource_cls is RawDatabase:
                # Raw Databases are automatically created when a Raw Table is created.
                continue
            required_by = {
                (required, path.relative_to(project_config_dir))
                for required, path in self._dependencies_by_required[(loader_cls, id_)]
            }
            self.warn(MissingDependencyWarning(loader_cls.resource_cls.__name__, id_, required_by))

    def _check_resource_exists_in_cdf(
        self, client: ToolkitClient, loader_cls: type[ResourceLoader], id_: Hashable
    ) -> bool:
        """Check is the resource exists in the CDF project. If there are any issues assume it does not exist."""
        if id_ in self.existing_resources_by_loader[loader_cls]:
            return True

        if loader_cls not in self.instantiated_loaders:
            self.instantiated_loaders[loader_cls] = loader_cls.create_loader(client)
        loader = self.instantiated_loaders[loader_cls]
        with contextlib.suppress(CogniteAPIError):
            retrieved = loader.retrieve([id_])
            if retrieved:
                self.existing_resources_by_loader[loader_cls].add(id_)
            return True
        return False

    def check_built_resource(
        self,
        parsed: dict[str, Any] | list[dict[str, Any]],
        loader: type[ResourceLoader],
        source: SourceLocation,
    ) -> tuple[WarningList[FileReadWarning], list[tuple[Hashable, str]]]:
        warning_list = WarningList[FileReadWarning]()

        if loader.yaml_cls:
            warning_list.extend(validate_resource_yaml_pydantic(parsed, loader.yaml_cls, source.path))

        is_dict_item = isinstance(parsed, dict)
        items = [parsed] if isinstance(parsed, dict) else parsed

        identifier_kind_pairs: list[tuple[Hashable, str]] = []
        for no, item in enumerate(items, 1):
            element_no = None if is_dict_item else no

            identifier: Any | None = None
            # Raw Tables and Raw Databases can have different loaders in the same file.
            item_loader = loader
            try:
                identifier = item_loader.get_id(item)
            except KeyError as error:
                if loader is RawTableLoader:
                    try:
                        identifier = RawDatabaseLoader.get_id(item)
                        item_loader = RawDatabaseLoader
                    except KeyError:
                        warning_list.append(
                            MissingRequiredIdentifierWarning(source.path, element_no, tuple(), error.args)
                        )
                else:
                    warning_list.append(MissingRequiredIdentifierWarning(source.path, element_no, tuple(), error.args))

            if identifier:
                identifier_kind_pairs.append((identifier, item_loader.kind))
                if first_seen := self._ids_by_resource_type[item_loader].get(identifier):
                    if isinstance(identifier, RawDatabase):
                        # RawDatabases are picked up from both RawTables and RawDatabases files. Note it is not possible
                        # to define a raw table without also defining the raw database. Thus, it is impossible to
                        # avoid duplicated RawDatabase warnings if you have multiple RawTables files.
                        continue
                    if first_seen.hash != source.hash:
                        warning_list.append(DuplicatedItemWarning(source.path, identifier, first_seen.path))
                else:
                    self._ids_by_resource_type[item_loader][identifier] = source

                try:
                    dependencies = list(item_loader.get_dependent_items(item))
                except (AttributeError, ValueError, TypeError, KeyError):
                    if element_no:
                        location = f" at element {element_no}"
                    else:
                        location = ""
                    raise ToolkitYAMLFormatError(
                        f"Error in {source.path.as_posix()}{location}. "
                        f"Failed to extract dependencies from {item_loader.kind}."
                    )
                for dependency in dependencies:
                    self._dependencies_by_required[dependency].append((identifier, source.path))

            api_spec = item_loader.safe_get_write_cls_parameter_spec()
            if api_spec is not None and item_loader.yaml_cls is None:
                # If we have a yaml cls this is already validated on the file as a whole.
                resource_warnings = validate_resource_yaml(parsed, api_spec, source.path, element_no)
                warning_list.extend(resource_warnings)

            data_set_warnings = validate_data_set_is_set(items, loader.resource_cls, source.path)
            warning_list.extend(data_set_warnings)

            item_warnings = item_loader.check_item(item, filepath=source.path, element_no=element_no)
            warning_list.extend(item_warnings)

        return warning_list, identifier_kind_pairs

    @staticmethod
    def _is_system_resource(resource_cls: type[ResourceLoader], id_: Hashable) -> bool:
        """System resources are deployed to all CDF project and should not be checked for dependencies."""
        if resource_cls is SpaceLoader and isinstance(id_, str) and id_.startswith("cdf_"):
            return True
        elif (
            resource_cls in {ContainerLoader, ViewLoader, DataModelLoader, NodeLoader}
            and hasattr(id_, "space")
            and id_.space.startswith("cdf_")
        ):
            return True
        return False
